module normal
!$ use omp_lib
  implicit none
contains

subroutine normal_cdf( x, a, b, cdf )
!*****************************************************************************80
!
!! NORMAL_CDF evaluates the Normal CDF.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    23 February 1999
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, real( kind = 8 ) X, the argument of the CDF.
!
!    Input, real( kind = 8 ) A, B, the parameters of the PDF.
!    0.0D+00 < B.
!    A is the mean, B is the standard deviation.
!
!    Output, real( kind = 8 ) CDF, the value of the CDF.
!
  real(8), intent(in) :: a
  real(8), intent(in) :: b
  real(8), intent(out) :: cdf
  real(8), intent(in) :: x
  real(8) y

	

  y =( x - a ) / b

  call normal_01_cdf( y, cdf )
	!write(*,*) 'in normal_cdf', a, b, x, y, cdf
  return
end subroutine normal_cdf

subroutine normal_01_cdf( x, cdf )

!*****************************************************************************80
!
!! NORMAL_01_CDF evaluates the Normal 01 CDF.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    10 February 1999
!
!  Author:
!
!    John Burkardt
!
!  Reference: 
!
!    AG Adams,
!    Algorithm 39, 
!    Areas Under the Normal Curve,
!    Computer Journal, 
!    Volume 12, pages 197-198, 1969.
!
!  Parameters:
!
!    Input, real( kind = 8 ) X, the argument of the CDF.
!
!    Output, real( kind = 8 ) CDF, the value of the CDF.
!
  implicit none

  real(8), parameter :: a1 = 0.398942280444D+00
  real(8), parameter :: a2 = 0.399903438504D+00
  real(8), parameter :: a3 = 5.75885480458D+00
  real(8), parameter :: a4 = 29.8213557808D+00
  real(8), parameter :: a5 = 2.62433121679D+00
  real(8), parameter :: a6 = 48.6959930692D+00
  real(8), parameter :: a7 = 5.92885724438D+00
  real(8), parameter :: b0 = 0.398942280385D+00
  real(8), parameter :: b1 = 3.8052D-08
  real(8), parameter :: b2 = 1.00000615302D+00
  real(8), parameter :: b3 = 3.98064794D-04
  real(8), parameter :: b4 = 1.98615381364D+00
  real(8), parameter :: b5 = 0.151679116635D+00
  real(8), parameter :: b6 = 5.29330324926D+00
  real(8), parameter :: b7 = 4.8385912808D+00
  real(8), parameter :: b8 = 15.1508972451D+00
  real(8), parameter :: b9 = 0.742380924027D+00
  real(8), parameter :: b10 = 30.789933034D+00
  real(8), parameter :: b11 = 3.99019417011D+00
  real(8) cdf
  real(8) q
  real(8) x
  real(8) y
!
!  |X| <= 1.28.
!
  if( abs( x ) <= 1.28D+00 ) then

    y = 0.5D+00 * x * x

    q = 0.5D+00 - abs( x ) *( a1 - a2 * y /( y + a3 - a4 /( y + a5 &
      + a6 /( y + a7 ) ) ) )
!
!  1.28 < |X| <= 12.7
!
  else if( abs( x ) <= 12.7D+00 ) then

    y = 0.5D+00 * x * x

    q = exp( - y ) * b0 /( abs( x ) - b1 &
      + b2 /( abs( x ) + b3 &
      + b4 /( abs( x ) - b5 &
      + b6 /( abs( x ) + b7 &
      - b8 /( abs( x ) + b9 &
      + b10 /( abs( x ) + b11 ) ) ) ) ) )
!
!  12.7 < |X|
!
  else

    q = 0.0D+00

  end if
!
!  Take account of negative X.
!
  if( x < 0.0D+00 ) then
    cdf = q
  else
    cdf = 1.0D+00 - q
  end if

  return
end subroutine normal_01_cdf

subroutine normal_cdf_inv( cdf, a, b, x )
!*****************************************************************************80
!
!! NORMAL_CDF_INV inverts the Normal CDF.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    23 February 1999
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, real( kind = 8 ) CDF, the value of the CDF.
!    0.0D+00 <= CDF <= 1.0.
!
!    Input, real( kind = 8 ) A, B, the parameters of the PDF.
!    0.0D+00 < B.
!
!    Output, real( kind = 8 ) X, the corresponding argument.
!

  real(8), intent(in) :: a
  real(8), intent(in) :: b
  real(8), intent(in) :: cdf
  real(8), intent(out) :: x
  real(8) x2
!write(*,*) 'input in normal_cdf', cdf

  if( cdf < 0.0D+00 .or. 1.0D+00 < cdf ) then
    write( *, '(a)' ) ' '
    write( *, '(a)' ) 'NORMAL_CDF_INV - Fatal error!'
    write( *, '(a)' ) '  CDF < 0 or 1 < CDF.'
    stop
  end if

  call normal_01_cdf_inv( cdf, x2 )

  x = a + b * x2

  return
end subroutine normal_cdf_inv

function r8poly_value( n, a, x )

!***********************************************************************
!
!! R8POLY_VALUE evaluates an R8POLY.
!
!  Discussion:
!
!    For sanity's sake, the value of N indicates the NUMBER of
!    coefficients, or more precisely, the ORDER of the polynomial,
!    rather than the DEGREE of the polynomial.  The two quantities
!    differ by 1, but cause a great deal of confusion.
!
!    Given N and A, the form of the polynomial is:
!
!      p(x) = a(1) + a(2) * x + ... + a(n-1) * x^(n-2) + a(n) * x^(n-1)
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    13 August 2004
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, integer N, the order of the polynomial.
!
!    Input, real( kind = 8 ) A(N), the coefficients of the polynomial.
!    A(1) is the constant term.
!
!    Input, real( kind = 8 ) X, the point at which the polynomial is
!    to be evaluated.
!
!    Output, real( kind = 8 ) R8POLY_VALUE, the value of the polynomial at X.
!
  implicit none

  integer n

  real(8) a(n)
  real(8) r8poly_value
  integer i
  real(8) x

  r8poly_value = 0.0D+00
  do i = n, 1, -1
    r8poly_value = r8poly_value * x + a(i)
  end do

  return
end function r8poly_value


subroutine normal_01_cdf_inv( p, x )
!*****************************************************************************80
!
!! NORMAL_01_CDF_INV inverts the standard normal CDF.
!
!  Discussion:
!
!    The result is accurate to about 1 part in 10**16.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    05 June 2007
!
!  Author:
!
!    Original FORTRAN77 version by Michael Wichura.
!    FORTRAN90 version by John Burkardt
!
!  Reference:
!
!    Michael Wichura,
!    Algorithm AS241:
!    The Percentage Points of the Normal Distribution,
!    Applied Statistics,
!    Volume 37, Number 3, pages 477-484, 1988.
!
!  Parameters:
!
!    Input, real( kind = 8 ) P, the value of the cumulative probability
!    densitity function.  0 < P < 1.  If P is outside this range, an 
!    "infinite" value will be returned. 
!
!    Output, real( kind = 8 ) X, the normal deviate value
!    with the property that the probability of a standard normal deviate being
!    less than or equal to the value is P.

  real(8), parameter, dimension( 8 ) :: a =(/ &
    3.3871328727963666080D+00, &
    1.3314166789178437745D+02, &
    1.9715909503065514427D+03, &
    1.3731693765509461125D+04, &
    4.5921953931549871457D+04, &
    6.7265770927008700853D+04, &
    3.3430575583588128105D+04, &
    2.5090809287301226727D+03 /)
  real(8), parameter, dimension( 8 ) :: b =(/ &
    1.0D+00, &
    4.2313330701600911252D+01, &
    6.8718700749205790830D+02, &
    5.3941960214247511077D+03, &
    2.1213794301586595867D+04, &
    3.9307895800092710610D+04, &
    2.8729085735721942674D+04, &
    5.2264952788528545610D+03 /)
  real(8), parameter, dimension( 8 ) :: c =(/ &
    1.42343711074968357734D+00, &
    4.63033784615654529590D+00, &
    5.76949722146069140550D+00, &
    3.64784832476320460504D+00, &
    1.27045825245236838258D+00, &
    2.41780725177450611770D-01, &
    2.27238449892691845833D-02, &
    7.74545014278341407640D-04 /)
  real(8), parameter :: const1 = 0.180625D+00
  real(8), parameter :: const2 = 1.6D+00
  !real(8) r8poly_value
  real(8), parameter, dimension( 8 ) :: d =(/ &
    1.0D+00, &
    2.05319162663775882187D+00, &
    1.67638483018380384940D+00, &
    6.89767334985100004550D-01, &
    1.48103976427480074590D-01, &
    1.51986665636164571966D-02, &
    5.47593808499534494600D-04, &
    1.05075007164441684324D-09 /)
  real(8), parameter, dimension( 8 ) :: e =(/ &
    6.65790464350110377720D+00, &
    5.46378491116411436990D+00, &
    1.78482653991729133580D+00, &
    2.96560571828504891230D-01, &
    2.65321895265761230930D-02, &
    1.24266094738807843860D-03, &
    2.71155556874348757815D-05, &
    2.01033439929228813265D-07 /)
  real(8), parameter, dimension( 8 ) :: f =(/ &
    1.0D+00, &
    5.99832206555887937690D-01, &
    1.36929880922735805310D-01, &
    1.48753612908506148525D-02, &
    7.86869131145613259100D-04, &
    1.84631831751005468180D-05, &
    1.42151175831644588870D-07, &
    2.04426310338993978564D-15 /)
  integer, parameter :: i4_8 = 8
  real(8) p
  real(8) q
  real(8) r
  real(8), parameter :: split1 = 0.425D+00
  real(8), parameter :: split2 = 5.0D+00
  real(8) x

  if( p <= 0.0D+00 ) then
    x = -huge( x )
    return
  end if

  if( 1.0D+00 <= p ) then
    x = huge( x )
    return
  end if

  q = p - 0.5D+00

  if( abs( q ) <= split1 ) then

    r = const1 - q * q
    x = q * r8poly_value( i4_8, a, r ) / r8poly_value( i4_8, b, r )

  else

    if( q < 0.0D+00 ) then
      r = p
    else
      r = 1.0D+00 - p
    end if

    if( r <= 0.0D+00 ) then

      x = huge( x )

    else

      r = sqrt( -log( r ) )

      if( r <= split2 ) then

        r = r - const2
        x = r8poly_value( i4_8, c, r ) / r8poly_value( i4_8, d, r )

      else

        r = r - split2
        x = r8poly_value( i4_8, e, r ) / r8poly_value( i4_8, f, r )

      end if

    end if

    if( q < 0.0D+00 ) then
      x = -x
    end if

  end if

  return
end subroutine normal_01_cdf_inv


end module normal
